"""
High-level access to amati functionality.
"""

import importlib
import json
import sys
from pathlib import Path

from jinja2 import Environment, FileSystemLoader
from loguru import logger
from pydantic import BaseModel, ValidationError

# pylint: disable=wrong-import-position

sys.path.insert(0, str(Path(__file__).parent.parent))
from amati._error_handler import handle_errors
from amati._logging import Log, Logger
from amati._resolve_forward_references import resolve_forward_references
from amati.file_handler import load_file

type JSONPrimitive = str | int | float | bool | None
type JSONArray = list["JSONValue"]
type JSONObject = dict[str, "JSONValue"]
type JSONValue = JSONPrimitive | JSONArray | JSONObject


def dispatch(data: JSONObject) -> tuple[BaseModel | None, list[JSONObject] | None]:
    """
    Returns the correct model for the passed spec

    Args:
        data: A dictionary representing an OpenAPI specification

    Returns:
        A pydantic model representing the API specification
    """

    version: JSONValue = data.get("openapi")

    if not isinstance(version, str):
        raise TypeError("A OpenAPI specification version must be a string.")

    if not version:
        raise TypeError("An OpenAPI Specfication must contain a version.")

    version_map: dict[str, str] = {
        "3.1.1": "311",
        "3.1.0": "311",
        "3.0.4": "304",
        "3.0.3": "304",
        "3.0.2": "304",
        "3.0.1": "304",
        "3.0.0": "304",
    }

    module = importlib.import_module(f"amati.validators.oas{version_map[version]}")

    resolve_forward_references(module)

    try:
        model = module.OpenAPIObject(**data)
    except ValidationError as e:
        return None, json.loads(e.json())

    return model, None


def check(original: JSONObject, validated: BaseModel) -> bool:
    """
    Runs a consistency check on the output of amati.
    Determines whether the validated model is the same as the
    originally provided API Specification

    Args:
        original: The dictionary representation of the original file
        validated: A Pydantic model representing the original file

    Returns:
        Whether original and validated are the same.
    """

    original_ = json.dumps(original, sort_keys=True)

    json_dump = validated.model_dump_json(exclude_unset=True, by_alias=True)
    new_ = json.dumps(json.loads(json_dump), sort_keys=True)

    return original_ == new_


def run(
    file_path: str | Path,
    consistency_check: bool = False,
    local: bool = False,
    html_report: bool = False,
):
    """
    Runs the full amati process on a specific specification file.

     * Parses the YAML or JSON specification, gunzipping if necessary.
     * Validates the specification.
     * Runs a consistency check on the ouput of the validation to verify
       that the output is identical to the input.
     * Stores any errors found during validation.

    Args:
        file_path: The specification to be validated
        consistency_check: Whether or not to verify the output against the input
    """

    spec = Path(file_path)

    data = load_file(spec)

    logs: list[Log] = []

    with Logger.context():
        result, errors = dispatch(data)
        logs.extend(Logger.logs)

    if errors or logs:

        handled_errors: list[JSONObject] = handle_errors(errors, logs)

        file_name = Path(Path(file_path).parts[-1])
        error_file = file_name.with_suffix(file_name.suffix + ".errors")
        error_path = spec.parent

        if local:
            error_path = Path(".amati")

            if not error_path.exists():
                error_path.mkdir()

        with open(
            error_path / error_file.with_suffix(error_file.suffix + ".json"),
            "w",
            encoding="utf-8",
        ) as f:
            f.write(json.dumps(handled_errors))

        if html_report:
            env = Environment(
                loader=FileSystemLoader(".")
            )  # Assumes template is in the same directory
            template = env.get_template("TEMPLATE.html")

            # Render the template with your data
            html_output = template.render(errors=handled_errors)

            # Save the output to a file
            with open(
                error_path / error_file.with_suffix(error_file.suffix + ".html"),
                "w",
                encoding="utf-8",
            ) as f:
                f.write(html_output)

    if result and consistency_check:
        return check(data, result)

    return True


def discover(spec: str, discover_dir: str = ".") -> list[Path]:
    """
    Finds OpenAPI Specification files to validate

    Args:
        spec: The path to a specific OpenAPI specification file.
        discover_dir: The directory to search through.
    Returns:
        A list of specifications to validate.
    """

    specs: list[Path] = []

    # If a spec is provided, check if it exists and erorr if not
    if spec:
        spec_path = Path(spec)

        if not spec_path.exists():
            raise FileNotFoundError(f"File {spec} does not exist.")

        if not spec_path.is_file():
            raise IsADirectoryError(f"{spec} is a directory, not a file.")

        specs.append(spec_path)

        # End early if we're not also trying to find files
        if not discover_dir:
            return specs

    if Path("openapi.json").exists():
        specs.append(Path("openapi.json"))

    if Path("openapi.yaml").exists():
        specs.append(Path("openapi.yaml"))

    if specs:
        return specs

    if discover_dir == ".":
        raise FileNotFoundError(
            "openapi.json or openapi.yaml can't be found, use --discover or --spec."
        )

    specs = specs + list(Path(discover_dir).glob("**/openapi.json"))
    specs = specs + list(Path(discover_dir).glob("**/openapi.yaml"))

    if not specs:
        raise FileNotFoundError(
            "openapi.json or openapi.yaml can't be found, use --spec."
        )

    return specs


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(
        prog="amati",
        description="""
        Tests whether a OpenAPI specification is valid. Will look an openapi.json
        or openapi.yaml file in the directory that amati is called from. If 
        --discover is set will search the directory tree. If the specification
        does not follow the naming recommendation the --spec switch should be
        used.
        
        Creates a file <filename>.errors.json alongside the original specification
        containing a JSON representation of all the errors.
        """,
    )

    parser.add_argument(
        "-s",
        "--spec",
        required=False,
        help="The specification to be parsed",
    )

    parser.add_argument(
        "--consistency-check",
        required=False,
        action="store_true",
        help="Runs a consistency check between the input specification and the"
        " parsed specification",
    )

    parser.add_argument(
        "-d",
        "--discover",
        required=False,
        default=".",
        help="Searches the specified directory tree for openapi.yaml or openapi.json.",
    )

    parser.add_argument(
        "--local",
        required=False,
        action="store_true",
        help="Store errors local to the caller in a file called <file-name>.errors.json"
        "; a .amati/ directory will be created.",
    )

    parser.add_argument(
        "--html-report",
        required=False,
        action="store_true",
        help="Creates an HTML report of the errors, called <file-name>.errors.html,"
        " alongside the original file or in a .amati/ directory if the --local switch"
        " is used",
    )

    args = parser.parse_args()

    logger.remove()  # Remove the default logger
    # Add a new logger that outputs to stderr with a specific format
    logger.add(sys.stderr, format="{time} | {level} | {message}")

    logger.info("Starting amati")

    try:
        specifications = discover(args.spec, args.discover)
    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.error(str(e))
        sys.exit(1)

    for specification in specifications:

        successful_check = False
        logger.info(f"Processing specification {specification}")

        try:
            successful_check = run(
                specification, args.consistency_check, args.local, args.html_report
            )
            logger.info(f"Specification {specification} processed successfully.")
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error(f"Error processing {specification}, {str(e)}")
            sys.exit(1)

        if args.consistency_check and successful_check:
            logger.info(f"Consistency check successful for {specification}")
        elif args.consistency_check:
            logger.info(f"Consistency check failed for {specification}")

    logger.info("Stopping amati.")



"""
Script to download the HTTP Authentication Schemes (ISO 9110) from IANA
"""

import json
import pathlib

import requests
from bs4 import BeautifulSoup

if __name__ == "__main__":
    response = requests.get(
        "https://www.iana.org/domains/root/db",
        timeout=20,
    )
    response.raise_for_status()  # Raise an exception for HTTP errors

    data: list[str] = []

    soup: BeautifulSoup = BeautifulSoup(response.text, "html.parser")
    table = soup.find("table", id="tld-table")

    for row in table.find_all("tr")[1:]:  # type: ignore # Skip the header row
        data.append(row.find_all("td")[0].text.strip())  # type: ignore

    write_to = pathlib.Path(__file__).parent.parent.resolve() / "amati/data/tlds.json"

    with write_to.open("w") as f:
        f.write(json.dumps(data))



"""
Script to download the HTTP Authentication Schemes (ISO 9110) from IANA
"""

import csv
import io
import json
import pathlib

import requests

if __name__ == "__main__":

    raw_data: dict[str, str] = {
        "application": "https://www.iana.org/assignments/media-types/application.csv",
        "audio": "https://www.iana.org/assignments/media-types/audio.csv",
        # "example":
        "font": "https://www.iana.org/assignments/media-types/font.csv",
        "haptics": "https://www.iana.org/assignments/media-types/haptics.csv",
        "image": "https://www.iana.org/assignments/media-types/image.csv",
        "message": "https://www.iana.org/assignments/media-types/message.csv",
        "model": "https://www.iana.org/assignments/media-types/model.csv",
        "multipart": "https://www.iana.org/assignments/media-types/multipart.csv",
        "text": "https://www.iana.org/assignments/media-types/text.csv",
        "video": "https://www.iana.org/assignments/media-types/video.csv",
    }

    # Example has no valid values at the moment
    data: dict[str, list[str]] = {"example": []}

    for registry, file in raw_data.items():
        response = requests.get(
            file,
            timeout=20,
        )
        response.raise_for_status()  # Raise an exception for HTTP errors

        reader = csv.DictReader(io.StringIO(response.text))

        data[registry] = []

        for row in reader:
            data[registry].append(row["Name"])

    write_to = (
        pathlib.Path(__file__).parent.parent.resolve() / "amati/data/media-types.json"
    )

    with write_to.open("w") as f:
        f.write(json.dumps(data))



"""
Script to download the Hypertext Transfer Protocol (HTTP) Status Code Registry from IANA
"""

import csv
import io
import json
import pathlib
import re

import requests

if __name__ == "__main__":
    response = requests.get(
        "https://www.iana.org/assignments/http-status-codes/http-status-codes-1.csv",
        timeout=20,
    )

    reader = csv.DictReader(io.StringIO(response.text))

    data: dict[str, str] = {}
    pattern = re.compile(r"^(\d{3})-*(\d{3})$")

    for row in reader:
        if match := pattern.match(row["Value"]):
            start, stop = match.groups()
            for value in range(int(start), int(stop) + 1):
                data[str(value)] = row["Description"]
        else:
            data[row["Value"]] = row["Description"]

    write_to = (
        pathlib.Path(__file__).parent.parent.resolve()
        / "amati/data/http-status-codes.json"
    )

    with write_to.open("w") as f:
        f.write(json.dumps(data))



"""
Script to download the SPDX licence list
"""

import json
import pathlib

import requests

if __name__ == "__main__":
    response = requests.get(
        "https://raw.githubusercontent.com/spdx/license-list-data/refs/heads/main/json/licenses.json",  # pylint: disable=ignore line-too-long
        timeout=20,
    )
    response.raise_for_status()  # Raise an exception for HTTP errors

    data = json.dumps(response.json())

    write_to = (
        pathlib.Path(__file__).parent.parent.resolve() / "amati/data/spdx-licences.json"
    )

    with write_to.open("w") as f:
        f.write(data)



"""
Validates the OpenAPI Specification version 3.1.1

Note that per https://spec.openapis.org/oas/v3.1.1.html#relative-references-in-api-description-uris  # pylint: disable=line-too-long

> URIs used as references within an OpenAPI Description, or to external documentation
> or other supplementary information such as a license, are resolved as identifiers,
> and described by this specification as URIs.

> Note that some URI fields are named url for historical reasons, but the descriptive
> text for those fields uses the correct “URI” terminology.

"""

import re
from typing import Any, ClassVar, Optional
from typing_extensions import Self

from jsonschema.exceptions import ValidationError as JSONVSchemeValidationError
from jsonschema.protocols import Validator as JSONSchemaValidator
from jsonschema.validators import validator_for  # type: ignore
from pydantic import (
    ConfigDict,
    Field,
    RootModel,
    model_validator,
)

from amati import AmatiValueError
from amati import model_validators as mv
from amati._logging import Logger
from amati.fields import (
    SPDXURL,
    URI,
    SPDXIdentifier,
)
from amati.fields.commonmark import CommonMark
from amati.fields.json import JSON
from amati.fields.oas import OpenAPI
from amati.fields.spdx_licences import VALID_LICENCES
from amati.validators.generic import GenericObject, allow_extra_fields
from amati.validators.oas304 import (
    CallbackObject,
    ContactObject,
    EncodingObject,
    ExampleObject,
    ExternalDocumentationObject,
    HeaderObject,
    LinkObject,
    PathItemObject,
    PathsObject,
    RequestBodyObject,
    ResponseObject,
    ResponsesObject,
)
from amati.validators.oas304 import SecuritySchemeObject as OAS30SecuritySchemeObject
from amati.validators.oas304 import (
    ServerObject,
    TagObject,
    XMLObject,
)

TITLE = "OpenAPI Specification v3.1.1"

# Convenience naming to ensure that it's clear what's happening.
# https://spec.openapis.org/oas/v3.1.1.html#specification-extensions
specification_extensions = allow_extra_fields


@specification_extensions("x-")
class LicenceObject(GenericObject):
    """
    A model representing the OpenAPI Specification licence object §4.8.4

    OAS uses the SPDX licence list.

    # SPECFIX: The URI is mutually exclusive of the identifier. I don't see
    the purpose of this; if the identifier is a SPDX Identifier where's the
    harm in also including the URI
    """

    name: str = Field(min_length=1)
    # What difference does Optional make here?
    identifier: Optional[SPDXIdentifier] = None
    url: Optional[URI] = None
    _reference_uri: ClassVar[str] = URI(
        "https://spec.openapis.org/oas/v3.1.1.html#license-object"
    )

    _not_url_and_identifier = mv.only_one_of(["url", "identifier"])

    @model_validator(mode="after")
    def check_uri_associated_with_identifier(self: Self) -> Self:
        """
        Validate that the URL matches the provided licence identifier.

        This validator checks if the URL is listed among the known URLs for the
        specified licence identifier.

        Returns:
            The validated licence object
        """
        # URI only - should warn if not SPDX
        if self.url:
            try:
                SPDXURL(self.url)
            except AmatiValueError:
                Logger.log(
                    {
                        "msg": f"{str(self.url)} is not a valid SPDX URL",
                        "type": "warning",
                        "loc": (self.__class__.__name__,),
                        "input": self.url,
                        "url": self._reference_uri,
                    }
                )

        # Both Identifier and URI, technically invalid, but should check if
        # consistent
        if (
            self.url
            and self.identifier
            and str(self.url) not in VALID_LICENCES[self.identifier]
        ):
            Logger.log(
                {
                    "msg": f"{self.url} is not associated with the identifier {self.identifier}",  # pylint: disable=line-too-long
                    "type": "warning",
                    "loc": (self.__class__.__name__,),
                    "input": self.model_dump_json(),
                    "url": self._reference_uri,
                }
            )

        return self


class ReferenceObject(GenericObject):
    """
    Validates the OpenAPI Specification reference object - §4.8.23

    Note, "URIs" can be prefixed with a hash; this is because if the
    representation of the referenced document is JSON or YAML, then
    the fragment identifier SHOULD be interpreted as a JSON-Pointer
    as per RFC6901.
    """

    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    ref: URI = Field(alias="$ref")
    summary: Optional[str]
    description: Optional[CommonMark]
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#reference-object"
    )


@specification_extensions("x-")
class InfoObject(GenericObject):
    """
    Validates the OpenAPI Specification info object - §4.8.2:
    """

    title: str
    summary: Optional[str] = None
    description: Optional[str | CommonMark] = None
    termsOfService: Optional[str] = None  # pylint: disable=invalid-name
    contact: Optional[ContactObject] = None
    license: Optional[LicenceObject] = None
    version: str
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.1.1.html#info-object"
    )


@specification_extensions("x-")
class DiscriminatorObject(GenericObject):
    """
    Validates the OpenAPI Specification object - §4.8.25
    """

    # FIXME: Need post processing to determine whether the property actually exists
    # FIXME: The component and schema objects need to check that this is being used
    # properly.
    propertyName: str
    mapping: Optional[dict[str, str | URI]] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#discriminator-object"
    )


@specification_extensions("x-")
class ServerVariableObject(GenericObject):
    """
    Validates the OpenAPI Specification server variable object - §4.8.6
    """

    enum: Optional[list[str]] = Field(None, min_length=1)
    default: str = Field(min_length=1)
    description: Optional[str | CommonMark] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#server-variable-object"
    )

    @model_validator(mode="after")
    def check_enum_default(self: Self) -> Self:
        """
        Validate that the default value is in the enum list.

        Returns:
            The validated server variable object
        """
        if self.enum is None:
            return self

        if self.default not in self.enum:
            Logger.log(
                {
                    "msg": f"The default value {self.default} is not in the enum list {self.enum}",  # pylint: disable=line-too-long
                    "type": "value_error",
                    "loc": (self.__class__.__name__,),
                    "input": {"default": self.default, "enum": self.enum},
                    "url": self._reference_uri,
                }
            )

        return self


@specification_extensions("x-")
class OperationObject(GenericObject):
    """Validates the OpenAPI Specification operation object - §4.8.10"""

    tags: Optional[list[str]] = None
    summary: Optional[str] = None
    description: Optional[str | CommonMark] = None
    externalDocs: Optional[ExternalDocumentationObject] = None
    operationId: Optional[str] = None
    parameters: Optional[list["ParameterObject | ReferenceObject"]] = None
    requestBody: Optional["RequestBodyObject | ReferenceObject"] = None
    responses: Optional["ResponsesObject"] = None
    callbacks: Optional[dict[str, "CallbackObject | ReferenceObject"]] = None
    deprecated: Optional[bool] = False
    security: Optional[list["SecurityRequirementObject"]] = None
    servers: Optional[list[ServerObject]] = None

    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#operation-object"
    )


PARAMETER_STYLES: set[str] = {
    "matrix",
    "label",
    "simple",
    "form",
    "spaceDelimited",
    "pipeDelimited",
    "deepObject",
}


@specification_extensions("x-")
class ParameterObject(GenericObject):
    """Validates the OpenAPI Specification parameter object - §4.8.11"""

    name: str
    in_: str = Field(alias="in")
    description: Optional[str | CommonMark] = None
    required: Optional[bool] = None
    deprecated: Optional[bool] = None
    allowEmptyValue: Optional[bool] = None
    style: Optional[str] = None
    explode: Optional[bool] = None
    allowReserved: Optional[bool] = None
    schema_: Optional["SchemaObject"] = Field(alias="schema")
    example: Optional[Any] = None
    examples: Optional[dict[str, "ExampleObject | ReferenceObject"]] = None
    content: Optional[dict[str, "MediaTypeObject"]] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#parameter-object"
    )

    _in_valid = mv.if_then(
        conditions={"in_": mv.UNKNOWN},
        consequences={"in_": ["query", "header", "path", "cookie"]},
    )
    _path_location_is_required = mv.if_then(
        conditions={"in_": "path"}, consequences={"required": True}
    )
    _empty_value_only_with_query = mv.if_then(
        conditions={"allowEmptyValue": mv.UNKNOWN},
        consequences={"in_": PARAMETER_STYLES ^ {"query"}},
    )
    _style_is_valid = mv.if_then(
        conditions={"style": mv.UNKNOWN}, consequences={"style": list(PARAMETER_STYLES)}
    )
    _reserved_only_with_query = mv.if_then(
        conditions={"allowReserved": mv.UNKNOWN},
        consequences={"in_": PARAMETER_STYLES ^ {"query"}},
    )
    _disallowed_if_schema = mv.if_then(
        conditions={"schema_": mv.UNKNOWN}, consequences={"content": None}
    )
    _disallowed_if_content = mv.if_then(
        conditions={"content": mv.UNKNOWN},
        consequences={
            "content": None,
            "style": None,
            "explode": None,
            "allowReserved": None,
            "schema_": None,
        },
    )


@specification_extensions("x-")
class MediaTypeObject(GenericObject):
    """
    Validates the OpenAPI Specification media type object - §4.8.14
    """

    schema_: Optional["SchemaObject"] = Field(alias="schema", default=None)
    # FIXME: Define example
    example: Optional[Any] = None
    examples: Optional[dict[str, ExampleObject | ReferenceObject]] = None
    encoding: Optional["EncodingObject"] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#media-type-object"
    )


class SchemaObject(GenericObject):
    """
    Schema Object as per OAS 3.1.1 specification (section 4.8.24)

    This model defines only the OpenAPI-specific fields explicitly.
    Standard JSON Schema fields are allowed via the 'extra' config
    and validated through jsonschema.
    """

    model_config = ConfigDict(
        populate_by_name=True,
        extra="allow",  # Allow all standard JSON Schema fields
    )

    # OpenAPI-specific fields not in standard JSON Schema
    nullable: Optional[bool] = None  # OAS 3.0 style nullable flag
    discriminator: Optional[DiscriminatorObject] = None  # Polymorphism support
    readOnly: Optional[bool] = None  # Declares property as read-only for requests
    writeOnly: Optional[bool] = None  # Declares property as write-only for responses
    xml: Optional[XMLObject] = None  # XML metadata
    externalDocs: Optional[ExternalDocumentationObject] = None  # External documentation
    example: Optional[Any] = None  # Example of schema
    examples: Optional[list[Any]] = None  # Examples of schema (OAS 3.1)
    deprecated: Optional[bool] = None  # Specifies schema is deprecated

    # JSON Schema fields that need special handling in OAS context
    ref: Optional[str] = Field(
        default=None, alias="$ref"
    )  # Reference to another schema

    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#schema-object"
    )

    @model_validator(mode="after")
    def validate_schema(self):
        """
        Use jsonschema to validate the model as a valid JSON Schema
        """
        schema_dict = self.model_dump(exclude_none=True, by_alias=True)

        # Handle OAS 3.1 specific validations

        # 1. Convert nullable to type array with null if needed
        if schema_dict.get("nullable") is True and "type" in schema_dict:
            type_val = schema_dict["type"]
            if isinstance(type_val, str) and type_val != "null":
                schema_dict["type"] = [type_val, "null"]
            elif isinstance(type_val, list) and "null" not in type_val:
                schema_dict["type"] = type_val + ["null"]

        # 2. Validate the schema structure using jsonschema's meta-schema
        # Get the right validator based on the declared $schema or default
        # to Draft 2020-12
        schema_version = schema_dict.get(
            "$schema", "https://json-schema.org/draft/2020-12/schema"
        )
        try:
            validator_cls: JSONSchemaValidator = validator_for(  # type: ignore
                {"$schema": schema_version}
            )
            meta_schema: JSON = validator_cls.META_SCHEMA  # type: ignore

            # This will validate the structure conforms to JSON Schema
            validator_cls(meta_schema).validate(schema_dict)  # type: ignore
        except JSONVSchemeValidationError as e:
            Logger.log(
                {
                    "msg": f"Invalid JSON Schema: {e.message}",
                    "type": "value_error",
                    "loc": (self.__class__.__name__,),
                    "input": schema_dict,
                    "url": self._reference_uri,
                }
            )

        return self


class SecuritySchemeObject(OAS30SecuritySchemeObject):
    """
    Validates the OpenAPI Security Scheme object - §4.8.27
    """

    _SECURITY_SCHEME_TYPES: ClassVar[set[str]] = {
        "apiKey",
        "http",
        "oauth2",
        "openIdConnect",
        "mutualTLS",
    }


type _Requirement = dict[str, list[str]]


# NB This is implemented as a RootModel as there are no pre-defined field names.
class SecurityRequirementObject(RootModel[list[_Requirement] | _Requirement]):
    """
    Validates the OpenAPI Specification security requirement object - §4.8.30:
    """

    # FIXME: The name must be a valid Security Scheme - need to use post-processing
    # FIXME If the security scheme is of type "oauth2" or "openIdConnect", then the
    # value must be a list
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.1.1.html#security-requirement-object"
    )


@specification_extensions("x-")
class ComponentsObject(GenericObject):
    """
    Validates the OpenAPI Specification components object - §4.8.7
    """

    schemas: Optional[dict[str, SchemaObject | ReferenceObject]] = None
    responses: Optional[dict[str, ResponseObject | ReferenceObject]] = None
    parameters: Optional[dict[str, ParameterObject | ReferenceObject]] = None
    examples: Optional[dict[str, ExampleObject | ReferenceObject]] = None
    requestBodies: Optional[dict[str, RequestBodyObject | ReferenceObject]] = None
    headers: Optional[dict[str, HeaderObject | ReferenceObject]] = None
    securitySchemes: Optional[dict[str, SecuritySchemeObject | ReferenceObject]] = None
    links: Optional[dict[str, LinkObject | ReferenceObject]] = None
    callbacks: Optional[dict[str, CallbackObject | ReferenceObject]] = None
    pathItems: Optional[dict[str, PathItemObject]] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.1.1.html#components-object"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_all_fields(
        cls, data: dict[str, dict[str, Any]]
    ) -> dict[str, dict[str, Any]]:
        """
        Validates the components object.

        Args:
            data: The data to validate.

        Returns:
            The validated components object.
        """

        pattern: str = r"^[a-zA-Z0-9\.\-_]+$"

        # Validate each field in the components object
        for field_name, value in data.items():
            if field_name.startswith("x-"):
                continue

            if not isinstance(value, dict):  # type: ignore
                raise ValueError(
                    f"Invalid type for '{field_name}': expected dict, got {type(value)}"
                )

            for key in value.keys():
                if not re.match(pattern, key):
                    raise ValueError(
                        f"Invalid key '{key}' in '{field_name}': must match pattern {pattern}"  # pylint: disable=line-too-long
                    )

        return data


@specification_extensions("x-")
class OpenAPIObject(GenericObject):
    """
    Validates the OpenAPI Specification object - §4.1
    """

    openapi: OpenAPI
    info: InfoObject
    jsonSchemaDialect: Optional[URI] = None
    servers: Optional[list[ServerObject]] = Field(default=[ServerObject(url=URI("/"))])
    paths: Optional[PathsObject] = None
    webhooks: Optional[dict[str, PathItemObject]] = None
    components: Optional[ComponentsObject] = None
    security: Optional[list[SecurityRequirementObject]] = None
    tags: Optional[list[TagObject]] = None
    externalDocs: Optional[ExternalDocumentationObject] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.1.1.html#openapi-object"
    )



"""
Validates the OpenAPI Specification version 3.1.1

Note that per https://spec.openapis.org/oas/v3.0.4.html#relative-references-in-api-description-uris  # pylint: disable=line-too-long

> URIs used as references within an OpenAPI Description, or to external documentation
> or other supplementary information such as a license, are resolved as identifiers,
> and described by this specification as URIs.

> Note that some URI fields are named url for historical reasons, but the descriptive
> text for those fields uses the correct “URI” terminology.

"""

import re
from typing import Any, ClassVar, Optional
from typing_extensions import Self

from jsonschema.exceptions import ValidationError as JSONVSchemeValidationError
from jsonschema.protocols import Validator as JSONSchemaValidator
from jsonschema.validators import validator_for  # type: ignore
from pydantic import (
    ConfigDict,
    Field,
    RootModel,
    ValidationError,
    field_validator,
    model_validator,
)

from amati import AmatiValueError
from amati import model_validators as mv
from amati._logging import Logger
from amati.fields import (
    URI,
    Email,
    HTTPAuthenticationScheme,
    HTTPStatusCode,
    MediaType,
    URIType,
    URIWithVariables,
)
from amati.fields.commonmark import CommonMark
from amati.fields.json import JSON
from amati.fields.oas import OpenAPI, RuntimeExpression
from amati.validators.generic import GenericObject, allow_extra_fields

type JSONPrimitive = str | int | float | bool | None
type JSONArray = list["JSONValue"]
type JSONObject = dict[str, "JSONValue"]
type JSONValue = JSONPrimitive | JSONArray | JSONObject


TITLE = "OpenAPI Specification v3.0.4"

# Convenience naming to ensure that it's clear what's happening.
# https://spec.openapis.org/oas/v3.0.4.html#specification-extensions
specification_extensions = allow_extra_fields


@specification_extensions("x-")
class ContactObject(GenericObject):
    """
    Validates the OpenAPI Specification contact object - §4.8.3
    """

    name: Optional[str] = None
    url: Optional[URI] = None
    email: Optional[Email] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.0.4.html#contact-object"
    )


@specification_extensions("x-")
class LicenceObject(GenericObject):
    """
    A model representing the OpenAPI Specification licence object §4.8.4
    """

    name: str = Field(min_length=1)
    url: Optional[URI] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#license-object"
    )


class ReferenceObject(GenericObject):
    """
    Validates the OpenAPI Specification reference object - §4.8.23

    Note, "URIs" can be prefixed with a hash; this is because if the
    representation of the referenced document is JSON or YAML, then
    the fragment identifier SHOULD be interpreted as a JSON-Pointer
    as per RFC6901.
    """

    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    ref: URI = Field(alias="$ref")
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#reference-object"
    )


@specification_extensions("x-")
class InfoObject(GenericObject):
    """
    Validates the OpenAPI Specification info object - §4.8.2:
    """

    title: str
    description: Optional[str | CommonMark] = None
    termsOfService: Optional[str] = None  # pylint: disable=invalid-name
    contact: Optional[ContactObject] = None
    license: Optional[LicenceObject] = None
    version: str
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.0.4.html#info-object"
    )


class DiscriminatorObject(GenericObject):
    """
    Validates the OpenAPI Specification object - §4.8.25
    """

    # FIXME: Need post processing to determine whether the property actually exists
    # FIXME: The component and schema objects need to check that this is being used
    # properly.
    propertyName: str
    mapping: Optional[dict[str, str | URI]] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#discriminator-object"
    )


@specification_extensions("x-")
class ExampleObject(GenericObject):
    """
    Validates the OpenAPI Specification example object - §4.8.19
    """

    summary: Optional[str] = None
    description: Optional[str | CommonMark] = None
    value: Optional[JSONValue] = None
    externalValue: Optional[URI] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#example-object"
    )

    _not_value_and_external_value = mv.only_one_of(
        ["value", "externalValue"], "warning"
    )


@specification_extensions("x-")
class ServerVariableObject(GenericObject):
    """
    Validates the OpenAPI Specification server variable object - §4.8.6
    """

    enum: Optional[list[str]] = Field(None, min_length=1)
    default: str = Field(min_length=1)
    description: Optional[str | CommonMark] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#server-variable-object"
    )

    @model_validator(mode="after")
    def check_enum_default(self: Self) -> Self:
        """
        Validate that the default value is in the enum list.

        Returns:
            The validated server variable object
        """
        if self.enum is None:
            return self

        if self.default not in self.enum:
            Logger.log(
                {
                    "msg": f"The default value {self.default} is not in the enum list {self.enum}",  # pylint: disable=line-too-long
                    "type": "warning",
                    "loc": (self.__class__.__name__,),
                    "input": {"default": self.default, "enum": self.enum},
                    "url": self._reference_uri,
                }
            )

        return self


@specification_extensions("x-")
class ServerObject(GenericObject):
    """
    Validates the OpenAPI Specification server object - §4.8.5
    """

    url: URIWithVariables | URI
    description: Optional[str | CommonMark] = None
    variables: Optional[dict[str, ServerVariableObject]] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#server-object"
    )


@specification_extensions("x-")
class ExternalDocumentationObject(GenericObject):
    """
    Validates the OpenAPI Specification external documentation object - §4.8.22
    """

    description: Optional[str | CommonMark] = None
    url: URI
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#external-documentation-object"
    )


# FIXME: Specification extensions should be "^x-", but the implementation
# doesn't play well with ConfigDict(extra="allow"). This is the only case
# so less important to change as the eventual logic is still correct.
@specification_extensions(".*")
class PathsObject(GenericObject):
    """Validates the OpenAPI Specification paths object - §4.8.8"""

    model_config = ConfigDict(extra="allow")

    @model_validator(mode="before")
    @classmethod
    def paths_are_uris(cls, data: Any) -> Any:
        """
        Validates that paths are valid URIs, it's allowed that they
        have variables, e.g. /pets or /pets/{petID}

        Special-case specification extensions, which are also allowed.
        """

        for field in data.keys():

            # Specification extensions
            if field.startswith("x-"):
                continue

            URIWithVariables(field)

        return data


@specification_extensions("x-")
class OperationObject(GenericObject):
    """Validates the OpenAPI Specification operation object - §4.8.10"""

    tags: Optional[list[str]] = None
    summary: Optional[str] = None
    description: Optional[str | CommonMark] = None
    externalDocs: Optional[ExternalDocumentationObject] = None
    operationId: Optional[str] = None
    parameters: Optional[list["ParameterObject | ReferenceObject"]] = None
    requestBody: Optional["RequestBodyObject | ReferenceObject"] = None
    responses: "ResponsesObject"
    callbacks: Optional[dict[str, "CallbackObject | ReferenceObject"]] = None
    deprecated: Optional[bool] = False
    security: Optional[list["SecurityRequirementObject"]] = None
    servers: Optional[list[ServerObject]] = None

    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#operation-object"
    )


PARAMETER_STYLES: set[str] = {
    "matrix",
    "label",
    "simple",
    "form",
    "spaceDelimited",
    "pipeDelimited",
    "deepObject",
}


@specification_extensions("x-")
class ParameterObject(GenericObject):
    """Validates the OpenAPI Specification parameter object - §4.8.11"""

    name: str
    in_: str = Field(alias="in")
    description: Optional[str | CommonMark] = None
    required: Optional[bool] = None
    deprecated: Optional[bool] = None
    allowEmptyValue: Optional[bool] = None
    style: Optional[str] = None
    explode: Optional[bool] = None
    allowReserved: Optional[bool] = None
    schema_: Optional["SchemaObject | ReferenceObject"] = Field(alias="schema")
    example: Optional[Any] = None
    examples: Optional[dict[str, "ExampleObject | ReferenceObject"]] = None
    content: Optional[dict[str, "MediaTypeObject"]] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#parameter-object"
    )

    _in_valid = mv.if_then(
        conditions={"in_": mv.UNKNOWN},
        consequences={"in_": ["query", "header", "path", "cookie"]},
    )
    _path_location_is_required = mv.if_then(
        conditions={"in_": "path"}, consequences={"required": True}
    )
    _empty_value_only_with_query = mv.if_then(
        conditions={"allowEmptyValue": mv.UNKNOWN},
        consequences={"in_": PARAMETER_STYLES ^ {"query"}},
    )
    _style_is_valid = mv.if_then(
        conditions={"style": mv.UNKNOWN}, consequences={"style": list(PARAMETER_STYLES)}
    )
    _reserved_only_with_query = mv.if_then(
        conditions={"allowReserved": mv.UNKNOWN},
        consequences={"in_": PARAMETER_STYLES ^ {"query"}},
    )
    _disallowed_if_schema = mv.if_then(
        conditions={"schema_": mv.UNKNOWN}, consequences={"content": None}
    )
    _disallowed_if_content = mv.if_then(
        conditions={"content": mv.UNKNOWN},
        consequences={
            "content": None,
            "style": None,
            "explode": None,
            "allowReserved": None,
            "schema_": None,
        },
    )


@specification_extensions("x-")
class RequestBodyObject(GenericObject):
    """
    Validates the OpenAPI Specification request body object - §4.8.13
    """

    description: Optional[CommonMark | str] = None
    content: dict[str, "MediaTypeObject"]
    required: Optional[bool] = False
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#request-body-object"
    )


@specification_extensions("x-")
class MediaTypeObject(GenericObject):
    """
    Validates the OpenAPI Specification media type object - §4.8.14
    """

    schema_: Optional["SchemaObject | ReferenceObject"] = Field(
        alias="schema", default=None
    )
    # FIXME: Define example
    example: Optional[Any] = None
    examples: Optional[dict[str, ExampleObject | ReferenceObject]] = None
    encoding: Optional["EncodingObject"] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#media-type-object"
    )


@specification_extensions("x-")
class EncodingObject(GenericObject):
    """
    Validates the OpenAPI Specification media type object - §4.8.15
    """

    contentType: Optional[str] = None
    headers: Optional[dict[str, "HeaderObject | ReferenceObject"]] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#encoding object-object"
    )

    @field_validator("contentType", mode="after")
    @classmethod
    def check_content_type(cls, value: str) -> str:
        """
        contentType is a comma-separated list of media types.
        Check that they are all valid

        raises: ValueError
        """

        for media_type in value.split(","):
            MediaType(media_type.strip())

        return value


type _ResponsesObjectReturnType = dict[str, "ReferenceObject | ResponseObject"]


@specification_extensions(".*")
class ResponsesObject(GenericObject):
    """
    Validates the OpenAPI Specification responses object - §4.8.16
    """

    model_config = ConfigDict(
        extra="allow",
    )

    default: Optional["ResponseObject | ReferenceObject"] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#responses-object"
    )

    @classmethod
    def _choose_model(
        cls, value: Any, field_name: str
    ) -> "ReferenceObject | ResponseObject":
        """
        Choose the model to use for validation based on the type of value.

        Args:
            value: The value to validate.

        Returns:
            The model class to use for validation.
        """

        message = f"{field_name} must be a ResponseObject or ReferenceObject, got {type(value)}"  # pylint: disable=line-too-long

        try:
            return ResponseObject.model_validate(value)
        except ValidationError:
            try:
                return ReferenceObject.model_validate(value)
            except ValidationError as e:
                raise ValueError(message, ResponsesObject._reference_uri) from e

    @model_validator(mode="before")
    @classmethod
    def validate_all_fields(cls, data: dict[str, Any]) -> _ResponsesObjectReturnType:
        """
        Validates the responses object.
        """

        validated_data: _ResponsesObjectReturnType = {}

        for field_name, value in data.items():

            # If the value is a specification extension, allow it
            if field_name.startswith("x-"):
                validated_data[field_name] = value
                continue

            # If the value is the fixed field, "default", allow it
            if field_name == "default":
                if isinstance(value, dict):
                    validated_data[field_name] = ResponsesObject._choose_model(
                        value, field_name
                    )
                continue

            # Otherwise, if the field appears like a valid HTTP status code or a range
            if re.match(r"^[1-5]([0-9]{2}|XX)+$", str(field_name)):

                # Double check and raise a value error if not
                HTTPStatusCode(field_name)

                # and validate as a ResponseObject or ReferenceObject
                validated_data[field_name] = ResponsesObject._choose_model(
                    value, field_name
                )

                continue

            # If the field is not a valid HTTP status code or "default"
            raise ValueError(f"Invalid type for numeric field '{field_name}'")

        return validated_data


@specification_extensions("x-")
class ResponseObject(GenericObject):
    """
    Validates the OpenAPI Specification response object - §4.8.17
    """

    description: str | CommonMark
    headers: Optional[dict[str, "HeaderObject | ReferenceObject"]] = None
    content: Optional[dict[str, MediaTypeObject]] = None
    links: Optional[dict[str, "LinkObject | ReferenceObject"]] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#response-object"
    )


@specification_extensions("x-")
class CallbackObject(GenericObject):
    """
    Validates the OpenAPI Specification callback object - §4.8.18
    """

    model_config = ConfigDict(extra="allow")

    # The keys are runtime expressions that resolve to a URL
    # The values are Response Objects or Reference Objects
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#callback-object"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_all_fields(cls, data: dict[str, Any]) -> dict[str, "PathItemObject"]:
        """
        Validates the callback object.
        """

        validated_data: dict[str, PathItemObject] = {}

        # Everything after a { but before a } should be runtime expression
        pattern: str = r"\{([^}]+)\}"

        for field_name, value in data.items():

            # If the value is a specification extension, allow it
            if field_name.startswith("x-"):
                validated_data[field_name] = PathItemObject.model_validate(value)
                continue

            # Either the field name is a runtime expression, so test this:
            try:
                RuntimeExpression(field_name)
                validated_data[field_name] = PathItemObject.model_validate(value)
                continue
            except AmatiValueError:
                pass

            # Or, the field name is a runtime expression embedded in a string
            # value per https://spec.openapis.org/oas/latest.html#examples-0
            matches = re.findall(pattern, field_name)

            for match in matches:
                try:
                    RuntimeExpression(match)
                except AmatiValueError as e:
                    raise AmatiValueError(
                        f"Invalid runtime expression '{match}' in field '{field_name}'",
                        CallbackObject._reference_uri,
                    ) from e

            if matches:
                validated_data[field_name] = PathItemObject.model_validate(value)
            else:
                # If the field does not contain a valid runtime expression
                raise ValueError(f"Invalid type for numeric field '{field_name}'")

        return validated_data


@specification_extensions("x-")
class TagObject(GenericObject):
    """
    Validates the OpenAPI Specification tag object - §4.8.22
    """

    name: str
    description: Optional[str | CommonMark] = None
    externalDocs: Optional[ExternalDocumentationObject] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#tag-object"
    )


@specification_extensions("x-")
class LinkObject(GenericObject):
    """
    Validates the OpenAPI Specification link object - §4.8.20
    """

    operationRef: Optional[URI] = None
    operationId: Optional[str] = None
    parameters: Optional[dict[str, RuntimeExpression | JSONValue]] = None
    requestBody: Optional[JSONValue | RuntimeExpression] = None
    description: Optional[str | CommonMark] = None
    server: Optional[ServerObject] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#link-object"
    )

    _not_operationref_and_operationid = mv.only_one_of(
        fields=["operationRef", "operationId"]
    )


@specification_extensions("x-")
class HeaderObject(GenericObject):
    """
    Validates the OpenAPI Specification link object - §4.8.20
    """

    # Common schema/content fields
    description: Optional[str | CommonMark] = None
    required: Optional[bool] = Field(default=False)
    deprecated: Optional[bool] = Field(default=False)

    # Schema fields
    style: Optional[str] = Field(default="simple")
    explode: Optional[bool] = Field(default=False)
    schema_: Optional["SchemaObject | ReferenceObject"] = Field(
        alias="schema", default=None
    )
    example: Optional[JSONValue] = None
    examples: Optional[dict[str, ExampleObject | ReferenceObject]] = None

    # Content fields
    content: Optional[dict[str, MediaTypeObject]] = None

    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#link-object"
    )

    _not_schema_and_content = mv.only_one_of(["schema_", "content"])


@specification_extensions("x-")
class XMLObject(GenericObject):
    """
    Validates the OpenAPI Specification object - §4.8.26
    """

    name: Optional[str] = None
    namespace: Optional[URI] = None
    prefix: Optional[str] = None
    attribute: Optional[bool] = Field(default=False)
    wrapped: Optional[bool] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#xml-object"
    )

    @field_validator("namespace", mode="after")
    @classmethod
    def _validate_namespace(cls, value: URI) -> URI:
        """
        Validates that the namespace is not a relative URI.
        """
        if value.type == URIType.RELATIVE:
            message = "XML namespace {value} cannot be a relative URI"
            Logger.log(
                {
                    "msg": message,
                    "type": "value_error",
                    "loc": (cls.__name__,),
                    "input": value,
                    "url": cls._reference_uri,
                }
            )

        return value


class SchemaObject(GenericObject):
    """
    Schema Object as per OAS 3.1.1 specification (section 4.8.24)

    This model defines only the OpenAPI-specific fields explicitly.
    Standard JSON Schema fields are allowed via the 'extra' config
    and validated through jsonschema.
    """

    model_config = ConfigDict(
        populate_by_name=True,
        extra="allow",  # Allow all standard JSON Schema fields
    )

    # OpenAPI-specific fields not in standard JSON Schema
    nullable: Optional[bool] = None  # OAS 3.0 style nullable flag
    discriminator: Optional[DiscriminatorObject] = None  # Polymorphism support
    readOnly: Optional[bool] = None  # Declares property as read-only for requests
    writeOnly: Optional[bool] = None  # Declares property as write-only for responses
    xml: Optional[XMLObject] = None  # XML metadata
    externalDocs: Optional[ExternalDocumentationObject] = None  # External documentation
    example: Optional[Any] = None  # Example of schema
    examples: Optional[list[Any]] = None  # Examples of schema (OAS 3.1)
    deprecated: Optional[bool] = None  # Specifies schema is deprecated

    # JSON Schema fields that need special handling in OAS context
    ref: Optional[str] = Field(
        default=None, alias="$ref"
    )  # Reference to another schema

    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#schema-object"
    )

    @model_validator(mode="after")
    def validate_schema(self):
        """
        Use jsonschema to validate the model as a valid JSON Schema
        """
        schema_dict = self.model_dump(exclude_none=True, by_alias=True)

        # Handle OAS 3.1 specific validations

        # 1. Convert nullable to type array with null if needed
        if schema_dict.get("nullable") is True and "type" in schema_dict:
            type_val = schema_dict["type"]
            if isinstance(type_val, str) and type_val != "null":
                schema_dict["type"] = [type_val, "null"]
            elif isinstance(type_val, list) and "null" not in type_val:
                schema_dict["type"] = type_val + ["null"]

        # 2. Validate the schema structure using jsonschema's meta-schema
        # Get the right validator based on the declared $schema or default
        # to Draft 2020-12
        schema_version = schema_dict.get(
            "$schema", "https://json-schema.org/draft/2020-12/schema"
        )
        try:
            validator_cls: JSONSchemaValidator = validator_for(  # type: ignore
                {"$schema": schema_version}
            )
            meta_schema: JSON = validator_cls.META_SCHEMA  # type: ignore

            # This will validate the structure conforms to JSON Schema
            validator_cls(meta_schema).validate(schema_dict)  # type: ignore
        except JSONVSchemeValidationError as e:
            Logger.log(
                {
                    "msg": f"Invalid JSON Schema: {e.message}",
                    "type": "value_error",
                    "loc": (self.__class__.__name__,),
                    "input": schema_dict,
                    "url": self._reference_uri,
                }
            )

        return self


OAUTH_FLOW_TYPES: set[str] = {
    "implicit",
    "authorizationCode",
    "clientCredentials",
    "password",
}


@specification_extensions("x-")
class OAuthFlowObject(GenericObject):
    """
    Validates the OpenAPI OAuth Flow object - §4.8.29
    """

    type: Optional[str] = None
    authorizationUrl: Optional[URI] = None
    tokenUrl: Optional[URI] = None
    refreshUrl: Optional[URI] = None
    scopes: dict[str, str] = {}
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#oauth-flow-object"
    )

    _implicit_has_authorization_url = mv.if_then(
        conditions={"type": "implicit"},
        consequences={"authorizationUrl": mv.UNKNOWN},
    )

    _token_url_not_implicit = mv.if_then(
        conditions={"tokenUrl": mv.UNKNOWN},
        consequences={"type": OAUTH_FLOW_TYPES ^ {"implicit"}},
    )

    _authorization_code_has_urls = mv.if_then(
        conditions={"type": "authorizationCode"},
        consequences={"authorizationUrl": mv.UNKNOWN, "tokenUrl": mv.UNKNOWN},
    )

    _authorization_url_not_credentials_password = mv.if_then(
        conditions={"authorizationUrl": mv.UNKNOWN},
        consequences={"type": OAUTH_FLOW_TYPES ^ {"clientCredentials", "password"}},
    )

    _client_credentials_has_token = mv.if_then(
        conditions={"type": "clientCredentials"},
        consequences={"tokenUrl": mv.UNKNOWN},
    )
    _password_has_token = mv.if_then(
        conditions={"type": "password"}, consequences={"tokenUrl": mv.UNKNOWN}
    )


@specification_extensions("-x")
class OAuthFlowsObject(GenericObject):
    """
    Validates the OpenAPI OAuth Flows object - §4.8.28

    SPECFIX: Not all of these should be optional as an OAuth2 workflow
    without any credentials will not do anything.
    """

    implicit: Optional[OAuthFlowObject] = None
    password: Optional[OAuthFlowObject] = None
    clientCredentials: Optional[OAuthFlowObject] = None
    authorizationCode: Optional[OAuthFlowObject] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#oauth-flow-object"
    )

    @model_validator(mode="before")
    @classmethod
    def _push_down_type(cls, data: Any) -> Any:
        """
        Adds the type of OAuth2 flow, e.g. implicit, password to the child
        OAuthFlowObject so that additional validation can be done on this object.
        """

        for k, v in data.items():

            if isinstance(v, OAuthFlowObject):
                raise NotImplementedError("Must pass a dict")

            if v:
                data[k]["type"] = k

        return data


class SecuritySchemeObject(GenericObject):
    """
    Validates the OpenAPI Security Scheme object - §4.8.27
    """

    type: str
    description: Optional[str | CommonMark] = None
    name: Optional[str] = None
    in_: Optional[str] = Field(default=None, alias="in")
    scheme: Optional[HTTPAuthenticationScheme] = None
    bearerFormat: Optional[str] = None
    flows: Optional[OAuthFlowsObject] = None
    openIdConnectUrl: Optional[URI] = None

    _SECURITY_SCHEME_TYPES: ClassVar[set[str]] = {
        "apiKey",
        "http",
        "oauth2",
        "openIdConnect",
    }

    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#security-scheme-object-0"
    )

    _type_in_enum = mv.if_then(
        conditions={"type": mv.UNKNOWN}, consequences={"type": _SECURITY_SCHEME_TYPES}
    )

    _apikey_has_name_and_in = mv.if_then(
        conditions={"type": "apiKey"},
        consequences={"name": mv.UNKNOWN, "in_": ("query", "header", "cookie")},
    )

    _http_has_scheme = mv.if_then(
        conditions={"type": "http"}, consequences={"scheme": mv.UNKNOWN}
    )

    _oauth2_has_flows = mv.if_then(
        conditions={"type": "oauth2"}, consequences={"flows": mv.UNKNOWN}
    )

    _open_id_connect_has_url = mv.if_then(
        conditions={"type": "openIdConnect"},
        consequences={"openIdConnectUrl": mv.UNKNOWN},
    )

    _flows_not_oauth2 = mv.if_then(
        conditions={"flows": None},
        consequences={"type": _SECURITY_SCHEME_TYPES ^ {"oauth2"}},
    )


type _Requirement = dict[str, list[str]]


# NB This is implemented as a RootModel as there are no pre-defined field names.
class SecurityRequirementObject(RootModel[list[_Requirement] | _Requirement]):
    """
    Validates the OpenAPI Specification security requirement object - §4.8.30:
    """

    # FIXME: The name must be a valid Security Scheme - need to use post-processing
    # FIXME If the security scheme is of type "oauth2" or "openIdConnect", then the
    # value must be a list For other security scheme types, the array MAY contain a
    # list of role names which are required for the execution
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.0.4.html#security-requirement-object"
    )


@specification_extensions("x-")
class PathItemObject(GenericObject):
    """Validates the OpenAPI Specification path item object - §4.8.9"""

    ref_: Optional[URI] = Field(alias="$ref", default=None)
    summary: Optional[str] = None
    description: Optional[str | CommonMark] = None
    get: Optional[OperationObject] = None
    put: Optional[OperationObject] = None
    post: Optional[OperationObject] = None
    delete: Optional[OperationObject] = None
    options: Optional[OperationObject] = None
    head: Optional[OperationObject] = None
    patch: Optional[OperationObject] = None
    trace: Optional[OperationObject] = None
    servers: Optional[list[ServerObject]] = None
    parameters: Optional[list[ParameterObject | ReferenceObject]] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#path-item-object"
    )


@specification_extensions("x-")
class ComponentsObject(GenericObject):
    """
    Validates the OpenAPI Specification components object - §4.8.7
    """

    schemas: Optional[dict[str, SchemaObject]] = None
    responses: Optional[dict[str, ResponseObject | ReferenceObject]] = None
    parameters: Optional[dict[str, ParameterObject | ReferenceObject]] = None
    examples: Optional[dict[str, ExampleObject | ReferenceObject]] = None
    requestBodies: Optional[dict[str, RequestBodyObject | ReferenceObject]] = None
    headers: Optional[dict[str, HeaderObject | ReferenceObject]] = None
    securitySchemes: Optional[dict[str, SecuritySchemeObject | ReferenceObject]] = None
    links: Optional[dict[str, LinkObject | ReferenceObject]] = None
    callbacks: Optional[dict[str, CallbackObject | ReferenceObject]] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/v3.0.4.html#components-object"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_all_fields(
        cls, data: dict[str, dict[str, Any]]
    ) -> dict[str, dict[str, Any]]:
        """
        Validates the components object.

        Args:
            data: The data to validate.

        Returns:
            The validated components object.
        """

        pattern: str = r"^[a-zA-Z0-9\.\-_]+$"

        # Validate each field in the components object
        for field_name, value in data.items():
            if field_name.startswith("x-"):
                continue

            if not isinstance(value, dict):  # type: ignore
                raise ValueError(
                    f"Invalid type for '{field_name}': expected dict, got {type(value)}"
                )

            for key in value.keys():
                if not re.match(pattern, key):
                    raise ValueError(
                        f"Invalid key '{key}' in '{field_name}': must match pattern {pattern}"  # pylint: disable=line-too-long
                    )

        return data


@specification_extensions("x-")
class OpenAPIObject(GenericObject):
    """
    Validates the OpenAPI Specification object - §4.1
    """

    openapi: OpenAPI
    info: InfoObject
    servers: Optional[list[ServerObject]] = Field(default=[ServerObject(url=URI("/"))])
    paths: PathsObject
    components: Optional[ComponentsObject] = None
    security: Optional[list[SecurityRequirementObject]] = None
    tags: Optional[list[TagObject]] = None
    externalDocs: Optional[ExternalDocumentationObject] = None
    _reference_uri: ClassVar[str] = (
        "https://spec.openapis.org/oas/3.0.4.html#openapi-object"
    )
